import pytest

from cedardb.client import _create_insert_query


def test_insert_query():
    table_name = "mytbl"
    query = _create_insert_query(table_name=table_name, row_count=2)
    assert query == 'insert into "mytbl" values (%s,%s)'

    query = _create_insert_query(
        table_name=table_name, row_count=2, columns=["col1", "col2"]
    )
    assert query == 'insert into "mytbl" (col1,col2) values (%s,%s)'

    query = _create_insert_query(table_name=table_name, row_count=2, placeholder="?")
    assert query == 'insert into "mytbl" values (?,?)'

    with pytest.raises(AttributeError):
        _create_insert_query(None, row_count=1)
        _create_insert_query("some", row_count=None)



import psycopg

from cedardb.response import QueryResponse


def _create_insert_query(
    table_name: str,
    row_count: int,
    columns: list[str] | None = None,
    placeholder: str = "%s",
) -> str:
    """Creates an insert query with value placeholders.

    Args:
        table_name: The name of the table.
        row_count: How many values will be inserted, for placeholder generation.

    Returns:
        The insert query.

    Examples:
        >>> _create_insert_query("tbl", 4, ['col1', 'col2'])
        'insert into "tbl" ("col1", "col2") values (%s,%s,%s,%s)'
        >>> _create_insert_query("tbl", 3)
        'insert into "tbl" values (%s, %s, %s)'

    """
    if not table_name or not row_count:
        raise AttributeError(
            "In order to generate an insert query we need the table name and row_count to be"
            f" valid, but table_name={table_name!r} and row_count={row_count!r} was passed"
        )
    placeholders = f"{placeholder}," * row_count
    columns = " " + "(" + ",".join(columns) + ")" + " " if columns else " "
    if row_count > 0:
        placeholders = placeholders[:-1]

    return f'insert into "{table_name}"{columns}values ({placeholders})'


class Client:
    """Represents a client connecting to CedarDB instance.

    Examples:
        >>> from cedardb import Client
        >>> client = Client('localhost', dbname='postgres', user='postgres', password='postgres')

    """

    def __init__(
        self, host: str, dbname: str, user: str, password: str, port: str = "5432"
    ):
        self.host = host
        self.dbname = dbname
        self.user = user
        self.password = password
        self.port = port

    @property
    def conn_string(self):
        """Returns a PostgreSQL connection string in key-values.

        Examples:
            >>> Client(host="localhost", dbname="testdb", user="admin", password="secret", port="1234")
            "host=localhost port=1234 dbname=testdb user=admin password=secret"
        """
        return (
            f"host={self.host}"
            f" port={self.port}"
            f" dbname={self.dbname}"
            f" user={self.user}"
            f" password={self.password}"
        )

    def get_conn(self) -> psycopg.connection.Connection:
        """Returns a psycopg connection"""
        return psycopg.connect(self.conn_string)

    def query(
        self, statement: str, raise_exception: bool = False, factory=None
    ) -> QueryResponse:
        """
        Runs a sql `statement` to the connection.

        Args:
            statement:
            raise_exception:
            factory:

        Returns:

        """
        with self.get_conn() as conn:
            with conn.cursor() as cur:
                try:
                    cur.execute(statement)
                    response = QueryResponse.from_cur(cur, factory=factory)
                except psycopg.errors.Error as e:
                    response = QueryResponse.from_error(e, cur.statusmessage)

                    if raise_exception:
                        raise e

                return response

    def insert_many(
        self, table_name: str, rows, raise_exception: bool = False
    ) -> QueryResponse:
        """
        Leverages pipelining to bulk insert.

        See https://cedardb.com/docs/clients/python/#bulk-loading

        Examples:
            >>> rows = [ (1, 'data'), (2, 'data2')]
            >>> client.insert_many('table_name', rows)
        """
        statements = _create_insert_query(table_name, len(rows[0]))

        with self.get_conn() as conn:
            with conn.cursor() as cur:
                try:
                    cur.executemany(statements, rows)
                    response = QueryResponse.from_cur(
                        cur,
                        original_statement_type="INSERT PIPELINE",
                        result_expected=False,
                    )
                except psycopg.errors.Error as e:
                    response = QueryResponse.from_error(e, cur.statusmessage)

                    if raise_exception:
                        raise e

                return response



from unittest import mock
import pytest

import psycopg

from cedardb.client import Client
from tests.conftest import psycopg_mocked_with

client = Client(
    host="localhost", dbname="testdb", user="admin", password="secret", port="1234"
)


def test_client_uri():
    expected_uri = (
        "host=localhost" " port=1234" " dbname=testdb" " user=admin" " password=secret"
    )

    assert client.conn_string == expected_uri


@psycopg_mocked_with(rows=[(1, None), (2, 1)], columns=["column1", "column2"])
def test_client_select_ok(psycopg_mock):
    with mock.patch("psycopg.connect", return_value=psycopg_mock):
        r = client.query("select * from table")

        assert r.ok is True
        assert r.rows
        assert r.columns == ["column1", "column2"]
        assert r.rows == [(1, None), (2, 1)]
        assert r.row_count == 2
        assert r.exception is None
        assert r.error_message is None


@psycopg_mocked_with(raise_error=psycopg.errors.SyntaxError("some_error"))
def test_client_select_ko(psycopg_mock):
    with mock.patch("psycopg.connect", return_value=psycopg_mock):
        r = client.query("selec * from table")

        assert r.ok is False
        assert r.rows == []
        assert r.columns == []
        assert r.row_count == -1

        assert isinstance(r.exception, psycopg.errors.SyntaxError)
        assert r.error_message == "some_error"


@psycopg_mocked_with(raise_error=psycopg.errors.SyntaxError("some_error"))
def test_client_select_ko(psycopg_mock):
    with mock.patch("psycopg.connect", return_value=psycopg_mock):
        with pytest.raises(psycopg.errors.SyntaxError):
            client.query("selec * from table", raise_exception=True)

        r = client.query("selec * from table", raise_exception=False)
        assert r.error_message == "some_error"
        assert isinstance(r.exception, psycopg.errors.Error)


@psycopg_mocked_with(rows=[], columns=[], rowcount=2)
def test_client_insert_many(psycopg_mock):
    with mock.patch("psycopg.connect", return_value=psycopg_mock):
        rows = ((1, "somedata"), (2, "somedata2"))
        r = client.insert_many("mytbl", rows)
        assert r.row_count == 2
        assert r.ok is True
        assert not r.rows
        assert not r.columns
        assert r.original_statement_type == "INSERT PIPELINE"


@psycopg_mocked_with(
    raise_error=psycopg.errors.SyntaxError("some_error"),
    side_effect_method="executemany",
)
def test_client_insert_many_ko(psycopg_mock):
    with mock.patch("psycopg.connect", return_value=psycopg_mock):
        rows = ((1, "somedata"), (2, "somedata2"))
        with pytest.raises(psycopg.errors.SyntaxError):
            client.insert_many("mytbl", rows, raise_exception=True)

        r = client.insert_many("mytbl", rows, raise_exception=False)

        assert r.ok is False
        assert r.error_message == "some_error"
        assert isinstance(r.exception, psycopg.errors.Error)



import dataclasses

from .types import Columns, Rows, ColumnCenter


def center_string(
    string: str, pad_length: int, direction: ColumnCenter = "left"
) -> str:
    if direction == "left":
        return string.ljust(pad_length)

    if direction == "right":
        return string.rjust(pad_length)

    return string.center(pad_length)


def print_table(columns: Columns, rows: Rows, column_center: ColumnCenter = "left"):
    # The width of every column, calculated as the max
    # length from the column name, or the biggest value.
    col_widths = []

    for i, column in enumerate(columns):
        # Get the biggest value (in character count) of the values from a given column.
        biggest_value_in_chars: int = 0
        for row in rows:
            # We assume that the index of the column in columns matches the row index inside rows.
            biggest_value_in_chars = max(len(str(row[i])), biggest_value_in_chars)

        col_widths.append(max(biggest_value_in_chars, len(column)))

    # Create separator.
    separator = "+" + "+".join("-" * (width + 2) for width in col_widths) + "+"

    # Create header.
    header = (
        "| "
        + " | ".join(
            f"{center_string(col, col_widths[i], column_center)}"
            for i, col in enumerate(columns)
        )
        + " |"
    )

    # Create rows.
    row_lines = [
        "| "
        + " | ".join(
            f"{center_string(str(row[i]), col_widths[i], column_center)}"
            for i in range(len(columns))
        )
        + " |"
        for row in rows
    ]

    table = "\n".join([separator, header, separator, *row_lines, separator])
    return table


@dataclasses.dataclass
class QueryResponse:
    """Response from a SQL statement request."""

    original_statement_type: str
    columns: list[str] = dataclasses.field(default_factory=list)
    rows: list = dataclasses.field(repr=False, default_factory=list)
    row_count: int = -1
    duration: float = 0

    exception: Exception = None
    error_message: str = None

    def __iter__(self):
        return iter(self.rows)

    @classmethod
    def from_error(
        cls, e: Exception, original_statement_type: str = None
    ) -> "QueryResponse":
        return cls(
            exception=e,
            error_message=str(e).replace("\n", " "),
            original_statement_type="",
        )

    @classmethod
    def from_cur(
        cls,
        cur,
        original_statement_type: str = None,
        result_expected: bool = True,
        factory=None,
    ):
        """
        If a result is expected it will try to call `fetchall`.

        Args:
            cur:
            result_expected:
            factory:

        Returns:

        """
        if result_expected and cur.rowcount > 0:
            if factory:
                rows = list(map(lambda row: factory(*row), cur.fetchall()))
            else:
                rows = cur.fetchall()
        else:
            rows = []

        if not original_statement_type and cur.statusmessage:
            original_statement_type = cur.statusmessage.split(" ")[0]

        return cls(
            original_statement_type=original_statement_type,
            row_count=cur.rowcount,
            columns=[col.name for col in cur.description] if cur.description else [],
            rows=rows,
        )

    def _parse_description(self, description: str):
        """Parses psycopg cur.description"""

    def as_table(self, max_rows=10) -> str:
        # Error message
        if not self.columns and self.error_message:
            return print_table(["error"], [[self.error_message]])

        # No columns nor row, typical in DDL and DML.
        if not self.columns and not self.error_message:
            return print_table([" "], [[self.row_count]])

        return print_table(self.columns, self.rows[:max_rows], "left")

    @property
    def ok(self) -> bool:
        """Returns whether the query resulted in an error."""
        return not self.exception
